{***********************************************************************}
{   RWord  Version 1.3    Random Word Generator                         }
{   Copyright  1989  Y.Fujisawa                                         }
{   Machine:    J-3100B  MS-DOS 2.11                                    }
{   Compiler:   TURBO-PASCAL 5.0                                        }
{   History:                                                            }
{       89- 3- 3    1.0     New                                         }
{       89- 3- 5    1.1     procedure ReadFile was modified             }
{       89- 3- 8    1.2     Word group 1..9                             }
{       89- 3-23    1.3     Display heap error & specify repeat         }
{***********************************************************************}

program  RWord;

const
    WordLength = 60;
    DelimiterChar = '/';
type
    WordAttrType = 1..9;
    WordPtr = ^WordRec;
    WordRec = record
                St: string[WordLength];
                Attr: WordAttrType;
                Next: WordPtr;
            end;
    VocabRec = record
                Num: integer;
                Ptr: WordPtr;
               end;

var
    Vocabulary: array[WordAttrType] of VocabRec;

procedure  WriteMessage;
begin
    Writeln('RWord Version 1.3  Date:1989-3-23   Copyright 1989 Y.Fujisawa');
    Writeln('Usage:  RWORD [-n] <FileName>       ( n: Repeat times )');
end;  { WriteMessage }


{$f+}
function  HeapErrorFunc(Size: word): integer;
{$f-}
begin
    Writeln('[܂');
    HeapErrorFunc := 0;
end;  { HeapErrorFunc }


procedure  ReadFile(FileName: string);
var
    TextFile: text;
    LineCount: integer;
    Line: string;

    procedure  OpenFile;
    begin
        if Pos('.',FileName) = 0 then  FileName := FileName+'.RWD';
        Assign(TextFile,FileName);
        {$i-}  Reset(TextFile);  {$i+}
        if IOResult <> 0 then  begin
            Writeln('File open error !');
            Halt;
        end;  { if }
    end;  { OpenFile }

    procedure  ReadLine;

        procedure  ErrorExit;
        begin
            Writeln('Error at Line ',LineCount:3,'  ',Line);
            Close(TextFile);
            Halt;
        end;   { ErrorExit }

        function  GetWordAttr(Ch: char): WordAttrType;
        begin
            case UpCase(Ch) of
                '1'..'9':   GetWordAttr := Ord(Ch)-Ord('0');
                'A' :       GetWordAttr := 1;
                'N' :       GetWordAttr := 2;
                else    ErrorExit;
            end;  { case }
        end;  { GetWordType }

    var
        p: byte;
        at: WordAttrType;
        NewWord: WordPtr;
    begin
        Readln(TextFile, Line);
        Inc(LineCount);
        p := Pos(DelimiterChar,Line);
        if (p > 0) and ( p < Length(Line) ) then  begin
            New(NewWord);
            at := GetWordAttr(Line[p+1]);
            with NewWord^ do  begin
                St := Copy(Line,1,p-1);
                Attr := at;
                with Vocabulary[at] do  begin
                    Next := Ptr;
                    Ptr := NewWord;
                    Inc(Num);
                end;  { with }
            end;  { with }
        end  { if }
        else  ErrorExit;
    end;  { ReadLine }

var
    i: integer;
begin
    LineCount := 0;
    for i := 1 to 9 do  begin
        with Vocabulary[i] do  begin
            Num := 0;
            Ptr := nil;
        end;  { with }
    end;  { for }
    OpenFile;
    while not Eof(TextFile) do ReadLine;
    Close(TextFile);
end;  { ReadFile }


function  RandomWord: string;

    function Scan(Ptr: WordPtr; n: integer): string;
    var
        i: integer;
    begin
        for i := 1 to n-1 do  Ptr := Ptr^.Next;
        Scan := Ptr^.St;
    end;  { Scan }

var
    s: string;
    at: WordAttrType;
begin
    s := '';
    for at := 1 to 9 do  begin
        with Vocabulary[at] do  begin
            if Num > 0 then  s := s+Scan(Ptr,Random(Num)+1);
        end;  { with }
    end;  { for }
    RandomWord := s;
end;  { RandomWord }


function  GetRepeat(St: string): integer;

    procedure  ErrorExit;
    begin
        Writeln('');
        Halt;
    end;  { ErrorExit }

var
    Num,Result: integer;
begin
    if St[1] = '-' then  begin
        Delete(St,1,1);
        Val(St,Num,Result);
        if Result = 0 then
            GetRepeat := Num
        else
            ErrorExit;
    end  { if }
    else  ErrorExit;
end;  { GetRepeat }


var
    RepeatTimes,i: integer;
begin
    HeapError := @HeapErrorFunc;
    Randomize;
    case ParamCount of
        1:  begin
                ReadFile(ParamStr(1));
                repeat
                    Write(RandomWord);
                    Readln;
                until  false;
            end;
        2:  begin
                RepeatTimes := GetRepeat(ParamStr(1));
                ReadFile(ParamStr(2));
                for i := 1 to RepeatTimes do  Writeln(RandomWord);
            end;
        else    WriteMessage;
    end;  { case }
end.
